using GraphicsManager.Enums;
using GraphicsManager.Interfaces;
using OpenTK.Graphics.OpenGL4;
using OpenTK.Mathematics;
using OpenTK.Windowing.Common;

namespace GraphicsManager.Objects.Core;

public abstract class ParentBase : Rectangle, IParent
{
	protected ParentBase(Texture? texture = null)
	: base(texture)
	{
	}
	
    public ControlList Controls { get; } = new();
    public int LastX { get; protected set; }
    public int LastY { get; protected set; }
    public int LastSX { get; protected set; }
    public int LastSY { get; protected set; }
    public int LastSW { get; protected set; }
    public int LastSH { get; protected set; }
    
    public Vector2i Position => Location;

    public void BaseParentResize()
    {
	    BlockDraw = true;
	    for (int i = 0; i < Controls.Length; i++)
	    {
		    if (!Controls[i].Loaded) continue;
		    bool top = (Controls[i].Anchor & ObjectAnchor.Top) == ObjectAnchor.Top;
		    bool left = (Controls[i].Anchor & ObjectAnchor.Left) == ObjectAnchor.Left;
		    bool right = (Controls[i].Anchor & ObjectAnchor.Right) == ObjectAnchor.Right;
		    bool bottom = (Controls[i].Anchor & ObjectAnchor.Bottom) == ObjectAnchor.Bottom;
		    if (!top && !bottom) { Controls[i].Anchor |= ObjectAnchor.Top; top = true; }
		    if (!left && !right) { Controls[i].Anchor |= ObjectAnchor.Left; left = true; }
		    
		    int lx, ly, sy, sx;
		    bool UpdateDistance = false, SizeChanged = false;
		    if ((Controls[i].Anchor & ObjectAnchor.PreventWidthChange) == ObjectAnchor.PreventWidthChange)
		    {
			    UpdateDistance = true;
			    lx = Controls[i].Location.X + ((Size.X - Controls[i].Distance.X - Controls[i].Size.X - Controls[i].Location.X) / 2);
			    sx = Controls[i].Size.X;
		    }
		    else
		    {
			    lx = (left ? Controls[i].Location.X : Size.X - Controls[i].Distance.X - Controls[i].Size.X);
			    sx = (right ? Size.X - Controls[i].Distance.X - lx : Controls[i].Size.X);
		    }
		    if ((Controls[i].Anchor & ObjectAnchor.PreventHeightChange) == ObjectAnchor.PreventHeightChange)
		    {
			    UpdateDistance = true;
			    ly = Controls[i].Location.Y + ((Size.Y - Controls[i].Distance.Y - Controls[i].Size.Y - Controls[i].Location.Y) / 2);
			    sy = Controls[i].Size.Y;
		    }
		    else
		    {
			    ly = (top ? Controls[i].Location.Y : Size.Y - Controls[i].Distance.Y - Controls[i].Size.Y);
			    sy = (bottom ? Size.Y - Controls[i].Distance.Y - ly : Controls[i].Size.Y);
		    }
		    
		    if (sx != Controls[i].Size.X || sy != Controls[i].Size.Y)
		    {
			    SizeChanged = true;
			    Controls[i].SetSize(sx, sy);
		    }
		    if (lx != Controls[i].Location.X || ly != Controls[i].Location.Y)
		    {
			    Controls[i].SetLocation(lx, ly);
		    }
		    if (UpdateDistance)
		    {
			    Controls[i].ForceDistanceUpdate(this);
		    }
		    if (SizeChanged && Controls[i] is IParent parent)
		    {
			    parent.ParentResize();
		    }
	    }
	    if (Parent is not null) Parent.TryDraw();
	    BlockDraw = false;
    }
    
    public virtual void ParentResize()
    {
	    BaseParentResize();
    }

    public override void UpdateOpenGLCords(bool draw = false)
    {
	    base.UpdateOpenGLCords(draw);
	    for (int i = 0; i < Controls.Length; i++)
	    {
		    Controls[i].UpdateOpenGLCords();
	    }
    }

    public override void SetLocation(int x, int y)
    {
	    base.SetLocation(x, y);
	    for (int i = 0; i < Controls.Length; i++)
	    {
		    Controls[i].UpdateOpenGLCords();
	    }
    }

    public override void SetSize(int w, int h)
    {
	    OldHeight = Size.Y;
	    OldWidth = Size.X;
	    base.SetSize(w, h);
    }

    public virtual void ReportSizeUpdate(IRenderObject Control)
    {
		
    }
    
    public override void Draw(int x, int y, int sx, int sy, int sw, int sh)
    {
	    if (Loaded && Visible && Location.X > 0 - Size.X && Location.Y > 0 - Size.Y)
	    {
		    DrawBase();
		    if (!UpdateScissorBox(ref x, ref y, ref sx, ref sy, ref sw, ref sh)) return;
		    GL.Scissor(sx, Window!.CS.Y - sy - sh, sw, sh);
		    for (int i = 0; i < Controls.Length; i++)
		    {
			    if (!Controls[i].Loaded)
			    {
				    BlockDraw = true;
				    Controls[i].LoadToParent(this, Window!);
				    if (this is IFlow flow) flow.ForceScrollUpdate();
				    BlockDraw = false;
			    }
			    if (Controls[i].Location.X >= Size.X ||
			        Controls[i].Location.Y >= Size.Y ||
			        Controls[i].Location.X + Controls[i].Size.X < 0 ||
			        Controls[i].Location.Y + Controls[i].Size.Y < 0) continue;
			    Controls[i].Draw(x,y,sx,sy,sw,sh);
			    if (Controls[i] is IParent) GL.Scissor(sx, Window!.CS.Y - sy - sh, sw, sh);
		    }
	    }
    }
    
    public bool CollectUpperFiles { get; set; }
    
    public virtual bool UpdateScissorBox(ref int x, ref int y, ref int sx, ref int sy, ref int sw, ref int sh, int xx = 0, int xy = 0, int lw = 0, int lh = 0, bool UpdateBoxInfo = true)
    {
	    int nw = Size.X - xx - lw,
		    nh = Size.Y - xy - lh,
		    nx = Location.X + xx,
		    ny = Location.Y + xy;
	    
	    if (nx - sx + x > sw || ny - sy + y > sh)
	    {
		    if (UpdateBoxInfo) NotifiNotVisible();
		    return false;
	    }

	    if (nx - sx + x > -1)
	    {
		    int add = nx - sx + x;
		    sx += add;
		    sw -= add;
		    if (nw < sw)
			    sw = nw;
	    }
	    else
	    {
		    if (nw + nx - sx + x < sw)
			    sw = nw + nx - sx + x;
	    }
	    if (ny - sy + y > -1)
	    {
		    int add = ny - sy + y;
		    sy += add;
		    sh -= add;
		    if (nh < sh)
			    sh = nh;
	    }
	    else
	    {
		    if (nh + ny - sy + y < sh)
			    sh = nh + ny - sy + y;
	    }

	    x += nx;
	    y += ny;

	    if (sw <= 0 || sh <= 0)
	    {
		    if (UpdateBoxInfo) NotifiNotVisible();
		    return false;
	    }

	    if (UpdateBoxInfo)
	    {
		    LastX = x;
		    LastY = y;
		    LastSX = sx;
		    LastSY = sy;
		    LastSW = sw;
		    LastSH = sh;
	    }
	    return true;
    }

    public virtual void ResizeDraw(int x, int y, int sx, int sy, int sw, int sh)
    {
	    if (Anchor == (ObjectAnchor.Top | ObjectAnchor.Left))
	    {
		    int nw = Math.Min(OldWidth, Size.X),
			    nh = Math.Min(OldHeight, Size.Y);
		    if (!UpdateScissorBox(ref x, ref y, ref sx, ref sy, ref sw, ref sh)) return;
		    Window!.CopyFromFrontToBack(x, y, Size.X, Size.Y, x, y);
	    }
	    else Draw(x,y,sx,sy,sw,sh);
    }
    
    public int OldWidth { get; protected set; }
    public int OldHeight { get; protected set; }
    
    public bool IgnoreVisForChildren { get; set; }
    

    public override void NotifiNotVisible()
    {
	    base.NotifiNotVisible();
	    for (int i = 0; i < Controls.Length; i++)
	    {
		    Controls[i].NotifiNotVisible();
	    }
    }

    public override void Clean()
    {
	    for (int i = 0; i < Controls.Length; i++)
	    {
		    Controls[i].Clean();
	    }
	    base.Clean();
    }
    
    public override void LoadToParent(IParent Parent, IWindow Window)
    {
	    if (Loaded) return;
	    bool PastBlockState = BlockDraw;
	    BlockDraw = true;
	    base.LoadToParent(Parent, Window);
	    for (int i = 0; i < Controls.Length; i++)
	    {
		    if (Controls[i].Loaded) continue;
		    Controls[i].LoadToParent(this, Window);
		    if (Controls[i] is IFlow flow) flow.ForceScrollUpdate();
	    }
	    BlockDraw = PastBlockState;
    }
    
    
	public float IntToWindow(float p, bool Y = false)
	{
		return Parent!.IntToWindow((Y ? this.Location.Y : Location.X), Y) + p;
	}
}