using System.Security.Cryptography;
using System.Timers;
using GraphicsManager.Enums;
using GraphicsManager.Interfaces;
using GraphicsManager.Objects.Core;
using OpenTK.Graphics.OpenGL;
using OpenTK.Mathematics;
using OpenTK.Windowing.Common;
using Timer = System.Timers.Timer;

namespace GraphicsManager.Objects;

public class FlowLayout : ParentBase, IFlow
{
    public FlowLayout(Texture? texture = null)
    :base(texture)
    {
	    AllowHoverFromBehind = true;
		t = new(50);
		t.Enabled = true;
		t.Elapsed += TOnElapsed;
		Controls.ControlAdded += ControlsOnControlAdded;
		Controls.ControlRemoved += ControlsOnControlRemoved;
	}
    
    public int SpaceBetweenObjects { get; set; } = 0;

    public void ForceScrollUpdate()
    {
	    ScrollSum = 0;
	    for (int i = 0; i < Controls.Length; i++)
	    {
		    ScrollSum += (uint)Controls[i].Size.Y + (uint)SpaceBetweenObjects;
	    }

	    ScrollSum -= (uint)SpaceBetweenObjects;
    }

    public double ScrollUpdatesInterval
    {
	    get => t.Interval;
	    set => t.Interval = value;
    }

    [Obsolete("Not Working", true)]
    public void ScrollToControl(IRenderObject Control)
	{
		if (Controls.Length < 1 || !Controls.Contains(Control)) throw new Exception("Control does not exist");
		if (Control.Location.Y >= 0 && (Control.Location.Y + Control.Size.Y) <= Size.Y) return;
		int index = Controls.IndexOf(Control);
		bool startbottom = false;
		if (Controls.Length - 1 - index < index) startbottom = true;
		int loc = 0;
		if ((Control.Location.Y + Control.Size.Y) >= Size.Y) loc = Size.Y - Control.Size.Y;
		int start = loc + Control.Size.Y;
		BlockDraw = true;
		bool top = false;
		if (startbottom)
		{
			for (int i = index + 1; i < Controls.Length; i++)
			{
				start += Controls[i].Size.Y;
			}

			for (int i = Controls.Length - 1; i >= 0; i--)
			{
				start -= Controls[i].Size.Y;
				Controls[i].Location = new(Controls[i].Location.X, start);
			}
		}
		else
		{
			for (int i = index; i >= 0; i--)
			{
				start -= Controls[i].Size.Y;
			}
			for (int i = 0; i < Controls.Length; i++)
			{
				start += Controls[i].Size.Y;
				Controls[i].Location = new(Controls[i].Location.X, start);
			}
		}

		uint c = ScrollSum;
		for (int i = Controls.Length - 1; i >= 0; i--)
		{
			if (Controls[i].Location.Y + Controls[i].Size.Y < 0 || Controls[i].Location.Y > Size.Y)
			{
				Controls[i].Visible = false;
			}

			c -= (uint)Controls[i].Size.Y;
			if (!top && Controls[i].Location.Y <= 0)
			{
				top = true;
				ScrollValueBackEnd = c - (uint)Controls[i].Location.Y;
			}
		}
		BlockDraw = false;
		if (Parent is not null) Parent.TryDraw();
	}

	public void ScrollToTop()
	{
		ScrollValue = 0;
	}

	public void ScrollToBottom()
	{
		ScrollValue = MaxScrollValue;
	}
	
	public override void ReportSizeUpdate(IRenderObject Control)
	{
		BlockDraw = true;
		int index = Controls.IndexOf(Control);
		if (index < Controls.Length - 1)
		{
			int dif = Control.Location.Y + Control.Size.Y - Controls[index + 1].Location.Y;
			Vector2i v3i = new(0, dif);
			for (int i = index + 1; i < Controls.Length; i++)
			{
				Controls[i].Location += v3i;
				if (Controls[i] is IParent p)
				{
					p.ParentResize();
				}
			}
		}
		else
		{
			ForceScrollUpdate();
		}
		BlockDraw = false;
	}

	private Task ControlsOnControlRemoved()
	{
		ScrollSum = 0;
		if (Controls.Length < 1) return Task.CompletedTask;
		if (Controls[0].Location.Y > 0) Controls[0].Location = new(0);
		bool top = false;
		for (int i = 1; i < Controls.Length; i++)
		{
			Controls[i].Location = new(0, Controls[Controls.Length - 1].Location.Y + Controls[Controls.Length - 1].Size.Y);
			if (Controls[i].Location.Y >= 0 && !top)
			{
				top = true;
				ScrollValueBackEnd = ScrollSum - (uint)Controls[i].Location.Y;
			}
			ScrollSum += (uint)Controls[i].Size.Y;
		}
		
		return Task.CompletedTask;
	}

	private Task ControlsOnControlAdded(int index, IRenderObject arg)
	{
		ScrollSum += (uint)arg.Size.Y;
		if (Controls.Length > 0)
		{
			if (index < Controls.Length)
			{
				arg.Location = Controls[index].Location;
				Controls[index].Location = new(arg.Location.X, Controls[index].Location.Y + arg.Size.Y);
				for (int i = index + 1; i < Controls.Length; i++)
				{
					Controls[i].Location = new(0, Controls[i - 1].Location.Y + Controls[i - 1].Size.Y);
				}
			}
			else arg.Location = new(0, Controls[Controls.Length - 1].Location.Y + Controls[Controls.Length - 1].Size.Y + SpaceBetweenObjects);
		}
		else arg.Location = new(0);
		if (arg is IParent par2)
		{
			for (int i = 0; i < par2.Controls.Length; i++)
			{
				par2.Controls[i].Distance = new(
					par2.Size.X - par2.Controls[i].Size.X - par2.Controls[i].Location.X,
					par2.Size.Y - par2.Controls[i].Size.Y - par2.Controls[i].Location.Y);
				
			}
		}
		if (arg is not ILabel) arg.Size = new(Size.X, arg.Size.Y);
		if (arg is IParent par)
		{
			for (int i = 0; i < par.Controls.Length; i++)
			{
				bool top = (par.Controls[i].Anchor & ObjectAnchor.Top) == ObjectAnchor.Top;
				bool left = (par.Controls[i].Anchor & ObjectAnchor.Left) == ObjectAnchor.Left;
				bool right = (par.Controls[i].Anchor & ObjectAnchor.Right) == ObjectAnchor.Right;
				bool bottom = (par.Controls[i].Anchor & ObjectAnchor.Bottom) == ObjectAnchor.Bottom;
				if (!top && !bottom) { par.Controls[i].Anchor |= ObjectAnchor.Top; top = true; }
				if (!left && !right) { par.Controls[i].Anchor |= ObjectAnchor.Left; left = true; }
				int lx = (left ? par.Controls[i].Location.X : Size.X - par.Controls[i].Distance.X - par.Controls[i].Size.X);
				int ly = (top ? par.Controls[i].Location.Y : Size.Y - par.Controls[i].Distance.Y - par.Controls[i].Size.Y);
				int sy = (bottom ? Size.Y - par.Controls[i].Distance.Y - ly : par.Controls[i].Size.Y);
				int sx = (right ? Size.X - par.Controls[i].Distance.X - lx : par.Controls[i].Size.X);
				par.Controls[i].Size = new(sx, sy);
				par.Controls[i].Location = new(lx, ly);
			}
		}
		if (arg is not ILabel) arg.Anchor = ObjectAnchor.Left | ObjectAnchor.Right | ObjectAnchor.Top;
		return Task.CompletedTask;
	}
	
	
	public override void ResizeDraw(int x, int y, int sx, int sy, int sw, int sh)
	{
		if (Anchor == (ObjectAnchor.Top | ObjectAnchor.Bottom | ObjectAnchor.Left))
		{
			if (OldHeight == Size.Y)
			{
				if (!UpdateScissorBox(ref x, ref y, ref sx, ref sy, ref sw, ref sh)) return;
				Window!.CopyFromFrontToBack(x, y, Size.X, Size.Y, x, y);
			}
			else if (OldHeight < Size.Y)
			{
				if (Visible && Location.X > 0 - Size.X && Location.Y > 0 - Size.Y)
				{
					if (!UpdateScissorBox(ref x, ref y, ref sx, ref sy, ref sw, ref sh)) return;
					Window!.CopyFromFrontToBack(x, y, OldWidth, OldHeight, x, y);
					int nsh = sh - OldHeight+1;
					if (nsh <= 0) return;
					GL.Scissor(sx, Window!.CS.Y - sy - sh, sw, nsh);
					DrawBase();
					GL.Scissor(sx, Window!.CS.Y - sy - sh, sw, sh);
		    
					for (int i = 0; i < Controls.Length; i++)
					{
						if (!Controls[i].Loaded)
						{
							BlockDraw = true;
							Controls[i].LoadToParent(this, Window!);
							if (this is IFlow flow) flow.ForceScrollUpdate();
							BlockDraw = false;
						}
						if (Controls[i].Location.X >= Size.X ||
						    Controls[i].Location.Y >= Size.Y ||
						    Controls[i].Location.X + Controls[i].Size.X < OldWidth ||
						    Controls[i].Location.Y + Controls[i].Size.Y < OldHeight) continue;
						
						
						if (Controls[i] is IParent pp)
						{
							pp.ResizeDraw(x,y,sx,sy,sw,sh);
							GL.Scissor(sx, Window!.CS.Y - sy - sh, sw, sh);
						}
						else
						{
							Controls[i].Draw(x,y,sx,sy,sw,sh);
						}
					}
				}
			}
			else
			{
				if (!UpdateScissorBox(ref x, ref y, ref sx, ref sy, ref sw, ref sh)) return;
				Window!.CopyFromFrontToBack(x, y, Size.X, Size.Y, x, y);
			}
		}
		else if (Anchor == (ObjectAnchor.Top | ObjectAnchor.Left))
		{
			if (!UpdateScissorBox(ref x, ref y, ref sx, ref sy, ref sw, ref sh)) return;
			Window!.CopyFromFrontToBack(x, y, Size.X, Size.Y, x, y);
		}
		else Draw(x,y,sx,sy,sw,sh);
	}

	public override void SetSize(int w, int h)
	{
		BlockDraw = true;
		base.SetSize(w, h);
		for (int i = 0; i < Controls.Length; i++)
		{
			Controls[i].SetSize(w, Controls[i].Size.Y);
		}
		if (Parent is not null) Parent.TryDraw();
		BlockDraw = false;
	}

	public override Vector2i Size
	{
		get => base.Size;
		set
		{
			SetSize(value.X, value.Y);
		}
	}
	
	public Vector2 MousePosition => Window!.MousePosition;

	public override void LoadToParent(IParent Parent, IWindow Window)
	{
		base.LoadToParent(Parent,Window);
		Window.MouseWheel += WindowOnMouseWheel;
	}

	Timer t;
	private int dist;
	
	private void WindowOnMouseWheel(MouseWheelEventArgs obj)
	{
		try
		{
			if (!MouseInside) return;
			dist += (int)(obj.OffsetY * HScrollPixels);
			if (!t.Enabled) t.Start();
		}
		catch (Exception e)
		{
			Console.WriteLine(e);
		}
		
	}

	public uint HScrollPixels { get; set; } = 30;

	private uint ScrollValueBackEnd { get; set; }

	public event Func<bool, uint, uint, Task>? FlowUpdate;

	public virtual uint ScrollValue
	{
		get => ScrollValueBackEnd;
		set
		{
			if (FlowUpdate is not null) FlowUpdate.Invoke(true, ScrollValueBackEnd, value);
			if (!Loaded || value > MaxScrollValue || value == ScrollValueBackEnd)
			{
				dist = 0;
				return;
			}
			BlockDraw = true;
			Vector2i v3i = new(0, (int)(value - ScrollValueBackEnd));
			for (int i = 0; i < Controls.Length; i++)
			{
				Controls[i].Location -= v3i;
			}

			BlockDraw = false;
			if (Parent is not null) Parent.TryDraw();
			dist = 0;
			ScrollValueBackEnd = value;
		}
	}
	
	public override void ParentResize()
	{
		BlockDraw = true;
		for (int i = 0; i < Controls.Length; i++)
		{
			if (Controls[i] is IParent parent)
			{
				parent.ParentResize();
			}
		}
		if (Parent is not null) Parent.TryDraw();
		BlockDraw = false;
	}
	
	public uint MaxScrollValue
	{
		get
		{
			if (ScrollSum > Size.Y) return (uint)(ScrollSum - Size.Y);
			return 0;
		}
	}

	private uint ss;

	public uint ScrollSum
	{
		get => ss;
		private set
		{
			ss = value;
		}
	}
	private bool scroleip;
	private void TOnElapsed(object? sender, ElapsedEventArgs e)
	{
		try
		{
			if (!scroleip && dist != 0)
			{
				scroleip = true;
				long n = ScrollValue - dist;
				if (n < 0) n = 0;
				if (n > MaxScrollValue) n = MaxScrollValue;
				Window!.Invoke(() =>ScrollValue = (uint)n);
				scroleip = false;
			}
			else
			{
				scroleip = false;
				t.Stop();
			}
		}
		catch (Exception exception)
		{
			Console.WriteLine(exception);
		}
	}
}