using System.Numerics;
using System.Runtime.CompilerServices;
using GraphicsManager.Enums;
using GraphicsManager.Objects.Core;
using OpenTK.Mathematics;

namespace GraphicsManager.Objects;

public class ProgressBar<TNumber> : ParentBase where TNumber : INumber<TNumber> 
{
    private Rectangle ProgresRct;

    public Color4 ProgressColor
    {
        get
        {
            return ProgresRct.BackgroundColor;
        }
        set
        {
            ProgresRct.BackgroundColor = value;
        }
    }
    
    public bool UpdateOnDraw { get; set; }
    
    public ProgressBar()
    {
        ProgresRct = new()
        {
            Location = new(DrawingGap.X, DrawingGap.Y),
            IgnoreHover = true,
            BackgroundColor = Color4.Green,
            Anchor = ObjectAnchor.All
        };
        Controls.Add(ProgresRct);
    }
    
    public ProgressBar(Texture Background, Texture Progress)
    :base(Background)
    {
        ProgresRct = new(Progress)
        {
            Location = new(DrawingGap.X, DrawingGap.Y),
            IgnoreHover = true,
            BackgroundColor = Color4.Green,
            TextureDisplay = TextureDisplay.ProgressHorizontalCenter,
            Anchor = ObjectAnchor.All
        };
        TextureDisplay = TextureDisplay.VerticalCenter;
        Controls.Add(ProgresRct);
    }
    
    public ProgressBar(Texture Background)
        :base(Background)
    {
        ProgresRct = new(Background)
        {
            Location = new(DrawingGap.X, DrawingGap.Y),
            IgnoreHover = true,
            BackgroundColor = Color4.Green,
            TextureDisplay = TextureDisplay.ProgressHorizontalCenter,
            Anchor = ObjectAnchor.All,
        };
        TextureDisplay = TextureDisplay.VerticalCenter;
        Controls.Add(ProgresRct);
    }

    public Shader InnerShader
    {
        get
        {
            return ProgresRct.Shader;
        }
        set
        {
            ProgresRct.Shader = value;
        }
    }

    public override Vector2i Size
    {
        get
        {
            return base.Size;
        }
        set
        {
            base.Size = value;
            if (!UpdateOnDraw) UpdateProgress();
        }
    }

    public override Vector2i Location
    {
        get
        {
            return base.Location;
        }
        set
        {
            Vector2i diff = value - base.Location;
            ProgresRct.Location -= diff;
            base.Location = value;
        }
    }

    public TNumber mpv = TNumber.One, pv = TNumber.Zero;
    public Vector2i ProgressGap = new(0);
    public Vector4i DrawingGap = new(0);

    public TNumber MaxProgressValue
    {
        get
        {
            return mpv;
        }
        set
        {
            mpv = value;
            if (!UpdateOnDraw) UpdateProgress();
        }
    }

    public bool DisallowAboveMax { get; set; } = true;

    public void AddAtomic(TNumber value)
    {
        if (DisallowAboveMax && value > MaxProgressValue) value = MaxProgressValue;
        if (typeof(TNumber) == typeof(ulong))
        {
            Interlocked.Add(ref Unsafe.As<TNumber, ulong>(ref pv), (ulong)(object)value!);
        }
        if (!UpdateOnDraw) UpdateProgress();
    }
    
    public TNumber ProgressValue
    {
        get
        {
            return pv;
        }
        set
        {
            if (DisallowAboveMax && value > MaxProgressValue) pv = MaxProgressValue;
            else pv = value;
            if (!UpdateOnDraw) UpdateProgress();
        }
    }

    public override void Draw(int x, int y, int sx, int sy, int sw, int sh)
    {
        if (UpdateOnDraw) UpdateProgress();
        base.Draw(x, y, sx,sy, sw, sh);
    }

    public TNumber GetValueFromX(int x)
    {
        if (x <= ProgressGap.X + DrawingGap.X) return TNumber.Zero;
        if (x >= Size.X - ProgressGap.Y - DrawingGap.Z) return MaxProgressValue;
        x -= DrawingGap.X; 
        x -= ProgressGap.X;
        double max_x = Size.X - ProgressGap.X - ProgressGap.Y - DrawingGap.X - DrawingGap.Z;
        double v = (x / max_x) * Convert.ToDouble(MaxProgressValue);
        Type tt = typeof(TNumber);
        if (tt == typeof(byte) || tt == typeof(sbyte) ||
            tt == typeof(ushort) || tt == typeof(short) ||
            tt == typeof(uint) || tt == typeof(int) ||
            tt == typeof(ulong) || tt == typeof(long))
        {
            return TNumber.CreateTruncating(Math.Round(v));
        }

        return TNumber.CreateTruncating(v);
    }
    
    public int GetInternalocation(TNumber Value, bool IgnoreStart = false, bool IgnoreEnd = false)
    {
        int x = ProgressGap.X;
        int xx = 0;
        if (IgnoreStart) x = 0;
        if (IgnoreEnd) xx = ProgressGap.Y;
        if (Value <= TNumber.Zero) return DrawingGap.X + x;
        if (Value >= MaxProgressValue) return Size.X - DrawingGap.X - DrawingGap.Z - xx;
        double percent = Convert.ToDouble(Value) / Convert.ToDouble(MaxProgressValue);
        int MaxPixels = Size.X - DrawingGap.X - DrawingGap.Z - x - ProgressGap.Y;
        return (int)(MaxPixels * percent) + ProgressGap.X;
    }

    public int GetParentLocation(TNumber Value, bool IgnoreStart = false, bool IgnoreEnd = false)
    {
        return GetInternalocation(Value, IgnoreStart, IgnoreEnd) + DrawingGap.X;
    }

    public void UpdateProgress()
    {
        var x = GetInternalocation(ProgressValue);
        if (ProgresRct.Location.X != DrawingGap.X || ProgresRct.Location.Y != DrawingGap.Y)
        {
            ProgresRct.Location = new(DrawingGap.X, DrawingGap.Y);
            ProgresRct.Size = new(x, Size.Y - DrawingGap.Y - DrawingGap.W);
        }
        else ProgresRct.Size = new(x, ProgresRct.Size.Y);
    }
}