﻿using GraphicsManager.Enums;
using GraphicsManager.Globals;
using GraphicsManager.Interfaces;
using GraphicsManager.Objects.Core;
using OpenTK.Graphics.OpenGL4;
using OpenTK.Mathematics;
using OpenTK.Windowing.Common;
using OpenTK.Windowing.Common.Input;
using OpenTK.Windowing.Desktop;
using OpenTK.Windowing.GraphicsLibraryFramework;

namespace GraphicsManager.Objects;

public class Rectangle : ITextureObject
{
	public static readonly Dictionary<IGLFWGraphicsContext, Shader> DefaultShader = new();
	public static readonly Dictionary<IGLFWGraphicsContext, Shader> DefaultAlphaShader = new();
	public static readonly Dictionary<IGLFWGraphicsContext, Shader> DefaultAlphaTextureShader = new();

	public ObjectAnchor Anchor { get; set; } = ObjectAnchor.Left | ObjectAnchor.Top;

	public Vector2i GetWindowLocation()
	{
		if (!Loaded) return Location;
		Vector2i loc = Location;
		IParent? p = Parent;
		while (p is not null)
		{
			loc += p.Position;
			p = p.Parent;
		}

		return loc;
	}
	
	public Vector2i GetParentLocation(IParent par)
	{
		if (!Loaded) return Location;
		Vector2i loc = Location;
		IParent? p = Parent;
		while (p is not null && p != par)
		{
			loc += p.Position;
			p = p.Parent;
		}

		return loc;
	}

	public List<Texture> Textures { get; set; } = new();
	public BetterContextMenu? ContextMenu { get; set; } = null;

	public Rectangle(Texture? texture = null)
	{
		if (texture is not null) Textures.Add(texture);
	}
	
	public virtual void Focus()
	{
		
	}
	public virtual void UnFocus()
	{
		
	}

	private Color4 _BackgroundColor { get; set; } = new(0, 0, 0, 255);

	public virtual Color4 BackgroundColor
	{
		get => _BackgroundColor;
		set
		{
			_BackgroundColor = value;
			TryDraw();
		}
	}

	private bool _Visible = true;
	public virtual bool Visible
	{
		get => _Visible;
		set
		{
			_Visible = value; 
			if (value) TryDraw();
			else
			{
				if (Parent is not null) Parent.TryDraw();
			}
		}
	}

	public virtual void SendKeyEvent(KeyboardKeyEventArgs KeyArgs)
	{
		
	}
	
	public virtual void SendClipEvent(string ClipString)
	{
		
	}

	public virtual void DrawBase()
	{
		if (Visible && Loaded && Location.X > 0 - Size.X && Location.Y > 0 - Size.Y)
		{
			IsVisible = true;
			if (!Window!.Context.IsCurrent) Window.Context.MakeCurrent();
			foreach (Texture tex in Textures)
			{
				tex.Use();
			}
			Shader.Use();
			Shader.SetMatrix4(Shader.GetUniformLocation("windowMatrix"), Window.WindowSizeMatrix);
			if (!Textures.Any() || Shader.Handle == DefaultAlphaShader[Window!.Context].Handle)
			{ 
				GL.Uniform4(Shader.GetUniformLocation("objColor"), BackgroundColor);
			}
			GL.BindVertexArray(ArrayObject);
			GL.DrawElements(PrimitiveType.Triangles, Indexs.Length, DrawElementsType.UnsignedShort, 0);
		}
	}

	public virtual void Draw(int x, int y, int sx, int sy, int sw, int sh)
	{
		DrawBase();
	}

	public Vector4i Margins { get; set; } = new();
	
	public virtual MouseCursor HoverMouse { get; set; } = MouseCursor.Default;
	public List<SubHitBox> SubHitBoxes { get; } = new();
	public event Func<IRenderObject, string[], Task>? FilesDroped;
	public void SendFilesEvent(string[] Files)
	{
		if (FilesDroped is not null) _ = FilesDroped.Invoke(this, Files);
	}

	public void ForceDistanceUpdate()
	{
		if (Parent is null) return;
		ForceDistanceUpdate(Parent);
	}
	
	public void ForceDistanceUpdate(IParent parent)
	{
		if (parent is IWindow w) Distance = new(w.CS.X - Size.X - Location.X, w.CS.Y - Size.Y - Location.Y);
		else Distance = new(parent.Size.X - Size.X - Location.X, parent.Size.Y - Size.Y - Location.Y);
	}

	public virtual void Clean()
	{
		GL.BindBuffer(BufferTarget.ArrayBuffer, 0);
		GL.DeleteBuffer(BufferObject);
		GL.DeleteBuffer(ElementBufferObject);
		GL.DeleteVertexArray(ArrayObject);
	}

	public virtual void LoadToParent(IParent Parent, IWindow Window)
	{
		if (Loaded) return;
		this.Parent = Parent;
		this.Window = Window;
		if (Shader is null)
		{
			if (!Textures.Any())Shader = DefaultShader[Window.Context];
			else Shader = Texture.TextureShader[Window.Context];
		}
		BufferObject = GL.GenBuffer();
		ArrayObject = GL.GenVertexArray();
		if (Textures.Any() && !Shader.ForTexture)
		{
			Shader = Texture.TextureShader[Window.Context];
		}
		ElementBufferObject = GL.GenBuffer();
		GL.BindBuffer(BufferTarget.ElementArrayBuffer, ElementBufferObject);
		GL.BufferData(BufferTarget.ElementArrayBuffer, Indexs.Length * sizeof(ushort), Indexs, Hint);
		Loaded = true;
		Window.MouseDown += Window_MouseDown;
		Window.FileDrop += WindowOnFileDrop;
		SetLocation(loc_.X, loc_.Y);
		if (Distance.X == 0 && Distance.Y == 0) ForceDistanceUpdate(Parent);
		if (WindowLoaded is not null) WindowLoaded.Invoke(this);
	}

	public void TransferOwners(IWindow w, IParent p)
	{
		if (Parent is not null)
		{
			Parent.Controls._internal.Remove(this);
			Window = w;
			Parent = p;
		}
		p.Controls.Add(this);
	}

	private void WindowOnFileDrop(FileDropEventArgs obj)
	{
		if (!MouseInside) return;
		if (FilesDroped is not null) _ = FilesDroped.Invoke(this, obj.FileNames);
	}

	private bool mi = false;

	public bool MouseInside
	{
		get => mi;
		set
		{
			if (Window is null) return;
			if (AllowHoverFromBehind)
			{
				mi = value;
				if (value && MouseEnter is not null)  MouseEnter.Invoke(this);
				if (!value && MouseLeave is not null)  MouseLeave.Invoke(this);
			}
			else
			{
				if (Window.HoveringControl == this && value)
				{
					mi = value;
					if (MouseEnter is not null) MouseEnter.Invoke(this);
				}

				if (Window.HoveringControl != this && !value)
				{
					mi = value;
					if (MouseLeave is not null) MouseLeave.Invoke(this);
				}
			}
		}
	}

	public IParent? Parent { get; private set; }
	public IWindow? Window { get; private set; }
	private bool ahfb = false;

	public bool AllowHoverFromBehind
	{
		get
		{
			return ahfb;
		}
		set
		{
			ahfb = value;
			if (!value)
			{
				if (MouseInside && Window is not null && Window.HoveringControl != this)
				{
					mi = false;
					if (MouseLeave is not null) MouseLeave.Invoke(this);
				}
			}
		}
	}

	private void Window_MouseDown(OpenTK.Windowing.Common.MouseButtonEventArgs e)
	{
		if (MouseInside && e.Button == MouseButton.Button1 && Clicked is not null) _ = Clicked.Invoke(this);
		if (MouseInside && e.Button == MouseButton.Button2 && ContextMenu is not null && Window!.HoveringControl == this) ContextMenu.ShowContext(Window!);
		if (e.Button != MouseButton.Button2 && ContextMenu is not null && ContextMenu.Visible) ContextMenu.HideContext(Window!);
	}

	~Rectangle()
	{
		GL.BindBuffer(BufferTarget.ArrayBuffer, 0);
		GL.DeleteBuffer(BufferObject);
		GL.DeleteBuffer(ElementBufferObject);
		GL.DeleteVertexArray(ArrayObject);
	}

	public Shader Shader { get; set; } = null!;
	public int ElementBufferObject { get; private set; }
	public int BufferObject { get; private set; }
	public bool IgnoreHover { get; set; }
	public int ArrayObject { get; private set; }
	private Vector2i size_ = new(), loc_;

	private bool UpdateOnCALL = true;

	public float[] Points
	{
		set
		{
			try
			{
				if (Loaded)
				{
					
					GL.BindBuffer(BufferTarget.ArrayBuffer, BufferObject);
					GL.BindVertexArray(ArrayObject);
					int NumberOfCords = 2;
					if (Textures.Any())
					{
						NumberOfCords = 4;
						GL.EnableVertexAttribArray(Textures.First().Location);
						GL.VertexAttribPointer(Textures.First().Location, 2, VertexAttribPointerType.Float, false, NumberOfCords * sizeof(float), 2 * sizeof(float));
					}
					GL.VertexAttribPointer(0, 2, VertexAttribPointerType.Float, false, NumberOfCords * sizeof(float), 0);
					GL.EnableVertexAttribArray(0);
					GL.BindBuffer(BufferTarget.ArrayBuffer, BufferObject);
					GL.BufferData(BufferTarget.ArrayBuffer, value.Length * sizeof(float), value, Hint);
					GL.BindVertexArray(ArrayObject);
					GL.BindBuffer(BufferTarget.ElementArrayBuffer, ElementBufferObject);
					GL.BufferData(BufferTarget.ElementArrayBuffer, Indexs.Length * sizeof(ushort), Indexs, Hint);
					if (Window is not null && Window.CanControleUpdate && Loaded && UpdateOnCALL) Parent!.TryDraw();
					
				}
			}
			catch (AccessViolationException v)
			{
				Console.WriteLine(v.Message);
			}
		}
	}

	private bool iv;

	public bool IsVisible
	{
		get
		{
			return iv && Visible;
		}
		private set
		{
			iv = value;
		}
	}

	public virtual void NotifiNotVisible()
	{
		IsVisible = false;
	}

	private bool bd = false;

	public bool BlockDraw
	{
		get
		{
			if (Parent is not null) return Parent.BlockDraw || bd;
			return bd;
		}
		set
		{
			bd = value;
		}
	}

	public bool BlendOverride;
	
	public virtual void TryDraw(int deapth = 0)
	{
		if (BlockDraw) return;
		if (Parent is null) return;
		if (Window is null) return;
		if (Window.UseLiveView && !IsVisible) return;
		if (!Window.UseSubFrames)
		{
			Parent.TryDraw();
			return;
		}
		if (Window.LastFrameIsResize)
		{
			Window.TryDraw();
			return;
		}
		if (BackgroundColor.A < 0.98 || !(BlendOverride || !Shader.CanBlend))
		{
			Parent.TryDraw(deapth+1);
			return;
		}

		Action a;
		int sf = Window.SubFrameCount;
		if (sf == 100) a = () =>
		{
			Window.TryDraw();
		};
		else a = () =>
		{
			if (Window.LogFrames) Console.WriteLine("Drawing Sub Frame: " + Window.Frame + "." + sf + " at depth of " + deapth);
			Window.Context.MakeCurrent();
			GL.Scissor(Parent.LastSX, Window!.CS.Y - Parent.LastSY - Parent.LastSH, Parent.LastSW, Parent.LastSH);
			Draw(Parent.LastX, Parent.LastY, Parent.LastSX, Parent.LastSY, Parent.LastSW, Parent.LastSH);
			Window.Context.SwapBuffers();
			Draw(Parent.LastX, Parent.LastY, Parent.LastSX, Parent.LastSY, Parent.LastSW, Parent.LastSH);
			Window.CheckParent(Window);
		};
			
		if (Window.InvokeRequired) Window.Invoke(a);
		else a.Invoke();
	}
	
	
	public event Func<IRenderObject, Task>? SizeChanged; 

	public ushort[] Indexs { get; set; } = new ushort[6] { 0, 1, 3, 1, 2, 3 };
	public BufferUsageHint Hint { get; set; } = BufferUsageHint.StaticDraw;
	public event Func<IRenderObject, Task>? Clicked;
	public bool Loaded { get; private set; }
	public event Func<IRenderObject, Task>? WindowLoaded;
	public event Func<IRenderObject, Task>? MouseEnter;
	public event Func<IRenderObject, Task>? MouseLeave;
	public object? Tag { get; set; }
	public Vector2i Distance { get; set; }

	private TextureDisplay td = TextureDisplay.Clamped;

	public TextureDisplay TextureDisplay
	{
		get
		{
			return td;
		}
		set
		{
			if (Textures.Count == 0) return;
			TextureDisplay old = td;
			td = value;
			float per = (float)Textures[0].RawSize!.Value.Y / Textures[0].RawSize!.Value.X;
			float diff = 0;
			switch (old)
			{
				default:
					switch (value)
					{
						case TextureDisplay.Clamped:
							Points = new float[]
							{
								saf.X,        laf.Y, Textures[0].MaxText.X,     Textures[0].MaxText.Y,
								saf.X,        saf.Y, Textures[0].MaxText.X,     0,
								laf.X,        saf.Y, 0,                         0,
								laf.X,        laf.Y, 0,                         Textures[0].MaxText.Y,
							};
							break;
						case TextureDisplay.HorizontalCenter or TextureDisplay.ProgressHorizontalCenter or TextureDisplay.TextureHorizontalCenter:
							
							if (Window is not null)
							{
								diff = Window.IntToWindow(Size.Y) + 1;
								if (value == TextureDisplay.TextureHorizontalCenter)
								{
									diff = Window.IntToWindow(Textures[0].RawSize!.Value.Y);
								}
							}
							Points = new float[]
							{
								saf.X,        laf.Y, Textures[0].MaxText.X,     Textures[0].MaxText.Y,	//4
								saf.X,        saf.Y, Textures[0].MaxText.X,     0,						//9
								laf.X,        saf.Y, 0,                         0,						//14
								laf.X,        laf.Y, 0,                         Textures[0].MaxText.Y,	//19
								saf.X - diff, laf.Y, Textures[0].MaxText.X-per, Textures[0].MaxText.Y,	//24
								saf.X - diff, saf.Y, Textures[0].MaxText.X-per, 0,						//29
								laf.X + diff, saf.Y, per,                       0,						//34
								laf.X + diff, laf.Y, per,                       Textures[0].MaxText.Y,	//39
							};
							Indexs = new ushort[]
							{
								0, 1, 4,
								1, 4, 5,
								4, 5, 6,
								4, 6, 7,
								3, 6, 7,
								2, 3, 6
							};
							break;
						
						case TextureDisplay.Center:
							per = (float)Textures[0].RawSize!.Value.X / 3;
							diff = 0;
							if (Window is not null)
							{
								if (Size.X > Textures[0].RawSize!.Value.X)
									diff = (Window.IntToWindow(Textures[0].RawSize!.Value.X) + 1) / 3;
								else
									diff = (Window.IntToWindow(Size.X) + 1) / 3;
							}
							Points = new float[]
							{
								saf.X,        laf.Y, Textures[0].MaxText.X,     Textures[0].MaxText.Y,	//4
								saf.X,        saf.Y, Textures[0].MaxText.X,     0,						//9
								laf.X,        saf.Y, 0,                         0,						//14
								laf.X,        laf.Y, 0,                         Textures[0].MaxText.Y,	//19
								saf.X - diff, laf.Y, Textures[0].MaxText.X-per, Textures[0].MaxText.Y,	//24
								saf.X - diff, saf.Y, Textures[0].MaxText.X-per, 0,						//29
								laf.X + diff, saf.Y, per,                       0,						//34
								laf.X + diff, laf.Y, per,                       Textures[0].MaxText.Y,	//39
								saf.X - diff, laf.Y, Textures[0].MaxText.X-per, Textures[0].MaxText.Y,	//24
								saf.X - diff, saf.Y, Textures[0].MaxText.X-per, 0,						//29
								laf.X + diff, saf.Y, per,                       0,						//34
								laf.X + diff, laf.Y, per,                       Textures[0].MaxText.Y,	//39
								saf.X - diff, laf.Y, Textures[0].MaxText.X-per, Textures[0].MaxText.Y,	//24
								saf.X - diff, saf.Y, Textures[0].MaxText.X-per, 0,						//29
								laf.X + diff, saf.Y, per,                       0,						//34
								laf.X + diff, laf.Y, per,                       Textures[0].MaxText.Y,	//39
							};
							Indexs = new ushort[]
							{
								0,  4,  11,
								4,  11, 12,
								4,  5,  12,
								5,  12, 13,
								1,  5,  6,
								5,  6,  13,
								10, 11, 12,
								10, 12, 15,
								12, 13, 14,
								12, 14, 15,
								6,  7,  13,
								7,  13, 14,
								3,  9,  10,
								9,  10, 15,
								8,  9,  14,
								9,  14, 15,
								2,  7,  8,
								7,  8,  14,
							};
							break;
					}
					break;
			}
			if (Parent is not null) Parent.TryDraw();
		}
	}
	
	private static void SetPoint(float[] temp, int start, float X, float Y, float Tx, float Ty)
	{
		temp[start] = X;
		temp[start + 1] = Y;
		temp[start + 2] = Tx;
		temp[start + 3] = Ty;
	}
	
	public virtual Vector2i Size
	{
		get
		{
			return size_;
		}
		set
		{
			SetSize(value.X, value.Y);
		}
	}

	public virtual void SetSize(int wh)
	{
		SetSize(wh, wh);
	}
	public virtual void SetSize(int w, int h)
	{
		size_.X = w;
		size_.Y = h;
		UpdateOpenGLCords(true);
		if (SizeChanged is not null) SizeChanged.Invoke(this);
		if (Parent is not null) Parent.ReportSizeUpdate(this);
	}
	
	public virtual void SetLocation(int xy)
	{
		SetLocation(xy, xy);
	}
	public virtual void SetLocation(int x, int y)
	{
		loc_.X = x;
		loc_.Y = y;
		UpdateOpenGLCords(true);
	}

	public virtual void UpdateOpenGLCords(bool draw = false)
	{
		if (Window is null || Parent is null) return;
		float[] temp = Textures.Count == 0
			? new float[8]
			: TextureDisplay switch
			{
				TextureDisplay.Clamped => new float[16],
				TextureDisplay.HorizontalCenter or TextureDisplay.ProgressHorizontalCenter => new float[32],
				TextureDisplay.Center => new float[64],
				_ => new float[16]
			};
		laf.X = Parent.IntToWindow(loc_.X);
		laf.Y = Parent.IntToWindow(loc_.Y, true);
		
		temp[(!Textures.Any() ? 4 : 8)] = laf.X;
		temp[(!Textures.Any() ? 6 : 12)] = laf.X;
		temp[1] = laf.Y;
		temp[(!Textures.Any() ? 7 : 13)] = laf.Y;
		saf.X = Parent.IntToWindow(Size.X + loc_.X);
		saf.Y = Parent.IntToWindow(Size.Y + loc_.Y, true);
		temp[0] = saf.X;
		temp[(!Textures.Any() ? 2 : 4)] = saf.X;
		temp[(!Textures.Any() ? 3 : 5)] = saf.Y;
		temp[(!Textures.Any() ? 5 : 9)] = saf.Y;
		if (Textures.Count > 0)
		{
			Vector2i s = Size;
			float diff = Window.IntToWindow(s.Y) + 1;
			float per = (float)Textures[0].RawSize!.Value.Y / Textures[0].RawSize!.Value.X;
			if (TextureDisplay == TextureDisplay.Center)
			{
				if (s.X > Textures[0].RawSize!.Value.X)
					diff = (Window.IntToWindow(Textures[0].RawSize!.Value.X) + 1) / 3;
				else
					diff = (Window.IntToWindow(s.X) + 1) / 3;
				per = Textures[0].MaxText.X / 3;
			}
			switch (TextureDisplay)
			{
				case TextureDisplay.Clamped:
					temp[2] = Textures[0].MaxText.X;
					temp[3] = Textures[0].MaxText.Y;
					temp[6] = Textures[0].MaxText.X;
					temp[15] = Textures[0].MaxText.Y;
					break;
				case TextureDisplay.HorizontalCenter:
					temp[2] = Textures[0].MaxText.X;
					temp[3] = Textures[0].MaxText.Y;
					temp[6] = Textures[0].MaxText.X;
					temp[15] = Textures[0].MaxText.Y;
					temp[16] = saf.X - diff; // top r
					temp[17] = laf.Y;
					temp[18] = Textures[0].MaxText.X - per;
					temp[19] = 1;
					
					temp[20] = saf.X - diff; // bot r
					temp[21] = saf.Y;
					temp[22] = Textures[0].MaxText.X - per;
					temp[23] = 0;
					
					temp[24] = laf.X + diff; // bot l
					temp[25] = saf.Y;
					temp[26] = per;
					temp[27] = 0;
					
					temp[28] = laf.X + diff; // top l
					temp[29] = laf.Y;
					temp[30] = per;
					temp[31] = 1;
					break;
				case TextureDisplay.Center:
					temp[2] = Textures[0].MaxText.X;
					temp[3] = Textures[0].MaxText.Y;
					temp[6] = Textures[0].MaxText.X;
					temp[15] = Textures[0].MaxText.Y;
					
					int dy, dx;
					if (s.Y > Textures[0].RawSize!.Value.Y)
						dy = Textures[0].RawSize!.Value.Y;
					else
						dy = s.Y;
					if (s.X > Textures[0].RawSize!.Value.X)
						dx = Textures[0].RawSize!.Value.X;
					else
						dx = s.X;

					if (dy > dx) dy = dx;
					else dx = dy;
					
					float diffy =  (dy + 1) / (float)3;
					diff =  (dx + 1) / (float)3;
					
					SetPoint(temp, 16, temp[0],        temp[1] + diffy, temp[2],                 temp[3] - (temp[3]*per));
					SetPoint(temp, 20, temp[0],        temp[5] - diffy, temp[2],                 temp[3] * per);
					SetPoint(temp, 24, temp[4] - diff, temp[5],         temp[2] - (temp[2]*per), temp[7]);
					SetPoint(temp, 28, temp[8] + diff, temp[5],         temp[2] * per,           temp[7]);
					SetPoint(temp, 32, temp[8],        temp[21],        temp[10],                temp[23]);
					SetPoint(temp, 36, temp[8],        temp[17],        temp[10],                temp[19]);
					SetPoint(temp, 40, temp[28],       temp[1],         temp[30],                temp[3]);
					SetPoint(temp, 44, temp[24],       temp[1],         temp[26],                temp[3]);
					SetPoint(temp, 48, temp[24],       temp[17],        temp[26],                temp[19]);
					SetPoint(temp, 52, temp[24],       temp[21],        temp[26],                temp[23]);
					SetPoint(temp, 56, temp[28],       temp[21],        temp[30],                temp[23]);
					SetPoint(temp, 60, temp[28],       temp[17],        temp[30],                temp[19]);
					break;
				case TextureDisplay.ProgressHorizontalCenter:
					if (s.X > s.Y)
					{
						if (s.X > s.Y + s.Y)
						{
							temp[2] = Textures[0].MaxText.X;
							temp[3] = Textures[0].MaxText.Y;
							temp[6] = Textures[0].MaxText.X;
							temp[15] = Textures[0].MaxText.Y;
							temp[16] = saf.X - diff; // top r
							temp[17] = laf.Y;
							temp[18] = Textures[0].MaxText.X - per;
							temp[19] = 1;
					
							temp[20] = saf.X - diff; // bot r
							temp[21] = saf.Y;
							temp[22] = 1 - per;
							temp[23] = 0;
					
							temp[24] = laf.X + diff; // bot l
							temp[25] = saf.Y;
							temp[26] = per;
							temp[27] = 0;
					
							temp[28] = laf.X + diff; // top l
							temp[29] = laf.Y;
							temp[30] = per;
							temp[31] = 1;
						}
						else
						{
							temp[2] = (0.666666f + (((s.X - s.Y) / (float)s.Y)* per)) * Textures[0].MaxText.X;
							temp[3] = Textures[0].MaxText.Y;
							temp[6] = temp[2];
							temp[15] = Textures[0].MaxText.Y;
							//start last 33% of texture
							temp[16] = saf.X; // top r
							temp[17] = laf.Y;
							temp[18] = temp[2];
							temp[19] = 1;

							temp[20] = saf.X; // bot r
							temp[21] = saf.Y;
							temp[22] = temp[2];
							temp[23] = 0;
					
							temp[24] = laf.X + diff; // bot l
							temp[25] = saf.Y;
							temp[26] = per;
							temp[27] = 0;
					
							temp[28] = laf.X + diff; // top l
							temp[29] = laf.Y;
							temp[30] = per;
							temp[31] = 1;
						}
					}
					else
					{
						//first 33% of texture
						temp[2] = ((s.X/(float)s.Y)*per);
						temp[3] = Textures[0].MaxText.Y;
						temp[6] = temp[2];
						temp[15] = Textures[0].MaxText.Y;
						temp[16] = saf.X; // top r
						temp[17] = laf.Y;
						temp[18] = temp[2];
						temp[19] = Textures[0].MaxText.Y;
					
						temp[20] = saf.X; // bot r
						temp[21] = saf.Y;
						temp[22] = temp[2];
						temp[23] = 0;
					
						temp[24] = saf.X; // bot l
						temp[25] = saf.Y;
						temp[26] = temp[2];
						temp[27] = 0;
					
						temp[28] = saf.X; // top l
						temp[29] = laf.Y;
						temp[30] = temp[2];
						temp[31] = Textures[0].MaxText.Y;
					}
					break;
			}
		}

		UpdateOnCALL = draw;
		Points = temp;
		UpdateOnCALL = true;
		temp = Array.Empty<float>();
	}
	
	public virtual Vector2i Location
	{
		get
		{
			return loc_;
		}
		set
		{
			SetLocation(value.X, value.Y);
		}
	}
	
	private Vector2 laf = new(), saf = new();

	public Vector2 LocationAsFloat { get { return laf; } }
	public Vector2 SizeAsFloat { get { return saf; } }
}
